/*
 * vupensux.c
 *
 * VUPEN (now Zerodium) is selling a futex-based 0day which manifests itself
 * as a race condition. It utilizes the priority-inheritance futex operations.
 * This kernel module mitigates that by hooking the futex system call and
 * forcing the PI futexes to return ENOSYS, logging the PID, name, and some
 * arguments of the calling process. It was designed with Tails in mind (based
 * on Debian stable), but should work on most Linux distros. Installing it on
 * Tails requires build-essential and linux-headers-$(uname -r).
 *
 * TODO When Tails supports livepatch, use it instead.
 *
 */

#include <asm/irq_vectors.h>
#include <asm/tlbflush.h>
#include <linux/futex.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>

#define NR64_futex 202
#define NR32_futex 240

#ifdef CONFIG_X86_64
#  define set_tables_rw(x) set_addr_rw((unsigned long)sys_call_table64); \
			   set_addr_rw((unsigned long)sys_call_table32)
#  define set_tables_ro(x) set_addr_ro((unsigned long)sys_call_table64); \
			   set_addr_ro((unsigned long)sys_call_table32)
#else
#  define set_tables_rw(x) set_addr_rw((unsigned long)sys_call_table32)
#  define set_tables_ro(x) set_addr_ro((unsigned long)sys_call_table32)
#endif

#if !(defined(ARCH_X86) || !defined(ARCH_X64))
#  error "Only x86(_64) architectures are supported"
#endif

#if defined(CONFIG_X86_64) && !defined(CONFIG_IA32_EMULATION)
#  error "IA32 emulation must be enabled in the 64bit build"
#endif

#if !defined(CONFIG_FUTEX)
#  error "This kernel doesn't have futex support. What's the point...?"
#endif

void **sys_call_table64;
void **sys_call_table32;

asmlinkage long (*original_futex64)(int *, int, int, uint32_t, int *, int);
asmlinkage long (*original_futex32)(int *, int, int, uint32_t, int *, int);

int  init_module(void);
void cleanup_module(void);

/*
 * PI (priority-inheritance) futexes are designed to handle situations where
 * priority inversion may become a problem. This occurs when a high-priority
 * task is blocked by a low-priority task holding a lock which it is waiting
 * to acquire, while medium-priority tasks preempt the low-priority task from
 * the CPU in a continuous loop. When this occurs, the high-priority task,
 * which should be allowed to get the lock immediately, ends up being blocked
 * indefinitely, resulting in priority inversion.
 *
 * Unfortunately, PI futexes are extremely complex and have resulted in
 * multiple high-profile vulnerabilities. There is even an extant 0day which is
 * sold by Zerodium which utilizes these futex operations via a race condition.
 * Luckily, pretty much nothing actually utilizes these operations, so it is
 * safe to outright block them. This function returns true if the argument
 * contains a priority-inheritance futex, and false if it does not.
 *
 * If any new PI futexes are created, add them to the switch statement below.
 */
static inline int is_pi_futex(const int op)
{
	const int cmd = op & FUTEX_CMD_MASK;

	switch (cmd) {
	case FUTEX_LOCK_PI:
	case FUTEX_UNLOCK_PI:
	case FUTEX_TRYLOCK_PI:
	case FUTEX_WAIT_REQUEUE_PI:
	case FUTEX_CMP_REQUEUE_PI:
		return true;
	default:
		return false;
	}
}

/*
 * Hook into the futex syscall and run our check against the arguments. If the
 * check passes, return the original, unhooked futex function. If the check
 * fails, return an error and print a message to the kernel log. In the future,
 * perhaps consider killing the offending process, or force it to coredump?
 *
 * Note that a TOCTOU can occur if the hook checks a pointer to userland memory
 * which is then operated on by the original syscall. Luckily for us, futex_op
 * is not even a pointer, so it does not pose a risk the way it is used. It is
 * important to keep this in mind in the future if any additional checks are
 * implemented on other arguments like *uaddr, though.
 */
#ifdef CONFIG_X86_64
asmlinkage long new_sys_futex64(int *uaddr, int futex_op, int val,
				uint32_t val2, int *uaddr2, int val3)
{
	if (is_pi_futex(futex_op)) {
		printk(KERN_WARNING "from %s[%d], attempted to call banned pi "
				    "futex with futex_op %d and val %d\n",
		       current->comm, task_pid_nr(current), futex_op, val);

		return -ENOSYS;
	}

	return original_futex64(uaddr, futex_op, val, val2, uaddr2, val3);
}
#endif

asmlinkage long new_sys_futex32(int *uaddr, int futex_op, int val, 
				uint32_t val2, int *uaddr2, int val3)
{
	if (is_pi_futex(futex_op)) {
		printk(KERN_WARNING "from %s[%d], attempted to call banned pi "
				    "futex with futex_op %d and val %d\n",
		       current->comm, task_pid_nr(current), futex_op, val);

		return -ENOSYS;
	}

	return original_futex32(uaddr, futex_op, val, val2, uaddr2, val3);
}

/*
 * The syscall tables haven't been writable since kernel 2.4. They've been
 * read-only ever since. In order to write to them, we'll have to change the
 * permissions on the pages. We'll mark them as read-only once we're done
 * messing with them, of course. Thanks to Corey Henderson for the code.
 *
 * Note that if this ever stops working, it will always be possible to disable
 * write-protect by disabling interrupts with cli, ANDing 0x10000 with cr0,
 * and re-enabling interrupts with sti. Re-enabling write-protect can be done
 * by ORing the same value with cr0 (ensuring that interrupts are disabled
 * during the process). The pax_open_kernel()/pax_close_kernel() functions
 * from grsecurity are not sufficient because they do not disable interrupts.
 */
static void set_addr_rw(const unsigned long addr)
{
	unsigned int level;
	pte_t *pte;

	pte = lookup_address(addr, &level);
	if (pte->pte &~ _PAGE_RW)
		pte->pte |= _PAGE_RW;

	local_flush_tlb();
}

static void set_addr_ro(const unsigned long addr)
{
	unsigned int level;
	pte_t *pte;

	pte = lookup_address(addr, &level);
	pte->pte = pte->pte &~_PAGE_RW;

	local_flush_tlb();
}

/*
 * Since Linux 2.6, sys_call_table and ia32_sys_call_table are no longer
 * exported. It's no fun to grep through System.map-$(uname -r). But we're in
 * kernelmode, so there's no reason we shouldn't be able to find the location
 * of the tables anyway. On a 32bit system, we'll get the syscall handler from
 * the 0x80 interrupt handler in the interrupt descriptor table. Thanks to
 * Alexei Lozovsky (ilammy) for all the syscall table-finding code.
 */
static u8 *get_32bit_handler(void)
{
	struct desc_ptr interrupt_descriptor_table;
	gate_desc *interrupt_gates;

	store_idt(&interrupt_descriptor_table);
	interrupt_gates = (gate_desc *)interrupt_descriptor_table.address;

	return (u8 *)gate_offset(interrupt_gates[IA32_SYSCALL_VECTOR]);
}

/*
 * On 64bit systems, syscalls are not called using the 0x80 interrupt, but
 * rather by using the dedicated sysenter/syscall instructions. The 64bit
 * syscall handler will be loaded into the LSTAR MSR.
 */
#ifdef CONFIG_X86_64
static u8 *get_64bit_handler(void)
{
	u64 system_call_entry;

	rdmsrl(MSR_LSTAR, system_call_entry);

	return (u8 *)system_call_entry;
}
#endif

/*
 * The syscall handlers execute the system call and store the return value.
 * The assembly source for the handlers are arch/x86/kernel/entry_32.S for the
 * native 32bit implementation, arch/x86/kernel/entry_64.S for the native 64bit
 * implementation, and arch/x86/kernel/ia32entry.S for the ia32 emulation
 * implementation. This is to explain the hex values in the following function.
 *
 * For the 64bit handler:
 *
 *	call *sys_call_table(,%rax,8) = call disp32(,%rax,8)  = FF 14 C5 ...
 *	movq %rax,RAX-ARGOFFSET(%rsp) = movq %rax,disp8(%rsp) = 48 89 44 24 ...
 *
 * And the 32bit handler:
 *
 *	call *sys_call_table(,%eax,4) = call disp32(,%eax,4)  = FF 14 85 ...
 *	movl %eax,PT_EAX(%esp)        = movl %eax,disp8(%esp) = 89 44 24 ...
 */
static unsigned long *find_table_ref(const u8 *code)
{
	unsigned int i;
	u32 offset;

	for (i = 0; i < 256; i++) {
#ifdef CONFIG_X86_64
		if (code[i + 0] == 0xff && code[i + 1] == 0x14 &&
		    code[i + 2] == 0xc5 && code[i + 7] == 0x48 &&
		    code[i + 8] == 0x89 && code[i + 9] == 0x44 &&
		    code[i +10] == 0x24) {
			offset = *((u32 *)&code[i + 3]);
			return (unsigned long *)(0xffffffff00000000 | offset);
		}
#else
		if (code[i + 0] == 0xff && code[i + 1] == 0x14 &&
		    code[i + 2] == 0x85 && code[i + 7] == 0x89 &&
		    code[i + 8] == 0x44 && code[i + 9] == 0x24) {
			offset = *((u32 *)&code[i + 3]);
			return (unsigned long *)offset;
		}
#endif
	}

	return NULL;
}

/*
 * This function runs as soon as the module is loaded. It tries to locate the
 * syscall tables, and exits in failure. On success, it backs up the address
 * of the original futex calls, and overwrites them with our hooks.
 */
int __init init_module(void)
{
	/* locate the 32bit and 64bit syscall tables */
#ifdef CONFIG_X86_64
	sys_call_table64 = (void *)find_table_ref(get_64bit_handler());
	if (sys_call_table64 == NULL) {
		printk(KERN_ERR "unable to locate syscall table\n");
		return 1;
	}
#endif
	sys_call_table32 = (void *)find_table_ref(get_32bit_handler());
	if (sys_call_table32 == NULL) {
		printk(KERN_ERR "unable to locate syscall table\n");
		return 1;
	}

	/* save a copy of the original futex addresses */
#ifdef CONFIG_X86_64
	original_futex64 = sys_call_table64[NR64_futex];
#endif
	original_futex32 = sys_call_table32[NR32_futex];

	/* overwrite the old syscalls with our new syscalls */
	set_tables_rw();
#ifdef CONFIG_X86_64
	sys_call_table64[NR64_futex] = new_sys_futex64;
#endif
	sys_call_table32[NR32_futex] = new_sys_futex32;
	set_tables_ro();

	printk(KERN_INFO "loaded vupensux module, pi futexes are disabled\n");
	return 0;
}

void __exit cleanup_module(void)
{
	/* restore the location of the original syscalls */
	set_tables_rw();
#ifdef CONFIG_X86_64
	sys_call_table64[NR64_futex] = original_futex64;
#endif
	sys_call_table32[NR32_futex] = original_futex32;
	set_tables_ro();

	printk(KERN_INFO "unloaded vupensux module, pi futexes are enabled\n");
}

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Mitigates a futex-based 0day sold by VUPEN (now Zerodium)");
